def create_metadata_table(TaXon_table_xlsx, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    TaXon_table_xslx_df = pd.read_excel(TaXon_table_xlsx)
    samples_list = TaXon_table_xslx_df.columns.tolist()[10:]
    samples_metadata_list = []

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(samples_list) + 1
    ############################################################################

    for sample in samples_list:
        sample_metadata = []
        sample_metadata.append(sample)
        for part in sample.split("_"):
            sample_metadata.append(part)
        samples_metadata_list.append(sample_metadata)

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    metadata_df = pd.DataFrame(samples_metadata_list)
    metadata_df.columns  = ["Samples"] + ["col_" + str(column) for column in metadata_df.columns.tolist()[1:]]
    metadata_df.to_excel(Meta_data_table_xlsx, index=False)

    closing_text = "Metadata table is found under:\n" + '/'.join(str(Meta_data_table_xlsx).split("/")[-4:])
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("meta data table", "analysis", TaXon_table_xlsx.name, Meta_data_table_xlsx.name, "nan", path_to_outdirs)



def alpha_diversity_scatter_plot(TaXon_table_xlsx, meta_data_to_test, width, heigth, scatter_size, path_to_outdirs, template):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path
    import webbrowser
    import plotly.graph_objects as go

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
    categories = Meta_data_table_df[meta_data_to_test].tolist()

    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.PopupError("The meta data has to differ between samples!", title="Error")
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        OTU_abundances_dict = {}
        samples_metadata_list = []

        # remove samples that do not fit the format
        for i, sample in enumerate(samples):
            meta_data = str(Meta_data_table_df.loc[i][meta_data_to_test])
            samples_metadata_list.append(meta_data)

        #################################
        # Calculate Alpha diversity measurements (= observed_otus)

        observed_otus_dict = {}
        samples_dict = {}

        for i, sample in enumerate(samples):
            observed_otus = len([taxon for taxon in TaXon_table_df[sample].values.tolist() if taxon != 0])
            category = samples_metadata_list[i]
            if category not in observed_otus_dict.keys():
                observed_otus_dict[category] = [observed_otus]
                samples_dict[category] = [sample]
            else:
                observed_otus_dict[category] = observed_otus_dict[category] + [observed_otus]
                samples_dict[category] = samples_dict[category] + [sample]

        ########################################
        # create the plot

        fig = go.Figure()
        for category in set(categories):
            fig.add_trace(go.Scatter(x=samples_dict[category], y=observed_otus_dict[category], mode='markers', name=category, marker=dict(size=int(scatter_size))))
        fig.update_layout(height=int(heigth), width=int(width), template=template, yaxis_title="# OTUs", showlegend=True)

        ## finish script
        output_pdf = Path(str(path_to_outdirs) + "/" + "Alpha_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_scatter_plot.pdf")
        output_html = Path(str(path_to_outdirs) + "/" + "Alpha_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_scatter_plot.html")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        sg.Popup("Alpha diversity estimate are found in", path_to_outdirs, "/Alpha_diversity/", title="Finished", keep_on_top=True)
        from taxontabletools.create_log import ttt_log
        ttt_log("alpha diversity scatter", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)

    else:
        sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)

def alpha_diversity_boxplot(TaXon_table_xlsx, meta_data_to_test, width, heigth, scatter_size, path_to_outdirs, template, theme):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path
    import webbrowser
    import plotly.graph_objects as go

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
    categories = Meta_data_table_df[meta_data_to_test].tolist()

    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.PopupError("The meta data has to differ between samples!", title="Error")
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        OTU_abundances_dict = {}
        samples_metadata_list = []

        # remove samples that do not fit the format
        for i, sample in enumerate(samples):
            meta_data = str(Meta_data_table_df.loc[i][meta_data_to_test])
            samples_metadata_list.append(meta_data)

        #################################
        # Calculate Alpha diversity measurements (= observed_otus)

        observed_otus_dict = {}

        for i, sample in enumerate(samples):
            observed_otus = len([taxon for taxon in TaXon_table_df[sample].values.tolist() if taxon != 0])
            category = samples_metadata_list[i]
            if category not in observed_otus_dict.keys():
                observed_otus_dict[category] = [observed_otus]
            else:
                observed_otus_dict[category] = observed_otus_dict[category] + [observed_otus]

        ########################################
        # create the plot

        fig = go.Figure()
        for category in set(categories):
            fig.add_trace(go.Box(y=observed_otus_dict[category], name=category, marker_color=color1, marker_line_color=color2, marker_line_width=0.2, opacity=opacity_value))
        fig.update_layout(height=int(heigth), width=int(width), template=template, yaxis_title="# OTUs", showlegend=False)

        ## finish script
        output_pdf = Path(str(path_to_outdirs) + "/" + "Alpha_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_boxplot.pdf")
        output_html = Path(str(path_to_outdirs) + "/" + "Alpha_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_boxplot.html")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        sg.Popup("Alpha diversity estimate are found in", path_to_outdirs, "/Alpha_diversity/", title="Finished", keep_on_top=True)
        from taxontabletools.create_log import ttt_log
        ttt_log("alpha diversity boxplot", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)

    else:
        sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)



def replicate_analysis(TaXon_table_xlsx, height, width, suffix_list, path_to_outdirs, template, theme):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path
    import matplotlib.pyplot as plt
    from matplotlib_venn import venn2
    from matplotlib_venn import venn3
    from matplotlib.pyplot import plot, ion, show
    import matplotlib.gridspec as gridspec
    import math, os, webbrowser
    import plotly.express as px

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    height = int(height)
    width = int(width)

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)

    sample_names = TaXon_table_df.columns[10:].tolist()
    OTUs = TaXon_table_df["IDs"].values.tolist()

    derep_sample_names_dict =  {}
    unique_sample_names_list = []
    replicates_dict = {}

    for sample in sample_names:
        sample_name = sample.split("_")[0:-1]
        unique_sample_names_list.append("_".join(sample_name))

    unique_sample_names_set = sorted(set(unique_sample_names_list))

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(unique_sample_names_set) + 1
    ############################################################################

    replicate_perc_shared_dict = {}

    replicate_analysis_name = Path(TaXon_table_xlsx).name.replace(".xlsx", "")
    dirName = Path(str(path_to_outdirs) + "/Replicate_analysis/" + replicate_analysis_name)
    if not os.path.exists(dirName):
        os.mkdir(dirName)

    for sample in unique_sample_names_set:

        for i, suffix in enumerate(suffix_list):
            replicates_dict["rep_" + str(i)] = sample + "_" + suffix_list[i]

        replicate_names_list = list(replicates_dict.values())

        try:
            # first collect the present OTUs from all replicates
            replicate_comparison_dict = {}
            for replicate in replicate_names_list:
                replicate_comparison_list = []
                for OTU in TaXon_table_df[["IDs", replicate]].values.tolist():
                    if OTU[1] != 0:
                        replicate_comparison_list.append(OTU[0])
                replicate_comparison_dict[replicate] = replicate_comparison_list

            # now compare them
            if len(replicate_comparison_dict.keys()) == 2:
                replicate_a = list(replicate_comparison_dict.keys())[0]
                replicate_b = list(replicate_comparison_dict.keys())[1]

                taxa_labels_a = sorted(replicate_comparison_dict[replicate_a])
                taxa_labels_b = sorted(replicate_comparison_dict[replicate_b])

                a_only = set(taxa_labels_a) - set(taxa_labels_b)
                len_a_only = len(a_only)
                b_only = set(taxa_labels_b) - set(taxa_labels_a)
                len_b_only = len(b_only)
                shared = set(taxa_labels_a) & set(taxa_labels_b)
                len_shared = len(shared)
                perc_shared = round(len_shared / (len_a_only + len_b_only + len_shared) * 100, 2)
                replicate_perc_shared_dict[sample] = perc_shared

                plt.figure(figsize=(10, 5))
                venn2(subsets = (len_a_only, len_b_only, len_shared), set_labels = (replicate_a, replicate_b))
                output_pdf = Path(str(dirName) + "/" + sample + ".pdf")
                plt.savefig(output_pdf, bbox_inches='tight')

            elif len(replicate_comparison_dict.keys()) == 3:

                replicate_a = list(replicate_comparison_dict.keys())[0]
                replicate_b = list(replicate_comparison_dict.keys())[1]
                replicate_c = list(replicate_comparison_dict.keys())[2]

                taxa_labels_a = sorted(replicate_comparison_dict[replicate_a])
                taxa_labels_b = sorted(replicate_comparison_dict[replicate_b])
                taxa_labels_c = sorted(replicate_comparison_dict[replicate_c])

                a_only = set(taxa_labels_a) - set(taxa_labels_b) - set(taxa_labels_c)
                len_a_only = len(a_only)
                b_only = set(taxa_labels_b) - set(taxa_labels_a) - set(taxa_labels_c)
                len_b_only = len(b_only)
                c_only = set(taxa_labels_c) - set(taxa_labels_a) - set(taxa_labels_b)
                len_c_only = len(c_only)

                shared_all = set(taxa_labels_a) & set(taxa_labels_b) & set(taxa_labels_c)
                len_shared_all = len(shared_all)
                shared_a_b = set(taxa_labels_a) & set(taxa_labels_b) - set(taxa_labels_c)
                len_shared_a_b = len(shared_a_b)
                shared_a_c = set(taxa_labels_a) & set(taxa_labels_c) - set(taxa_labels_b)
                len_shared_a_c = len(shared_a_c)
                shared_b_c = set(taxa_labels_b) & set(taxa_labels_c) - set(taxa_labels_a)
                len_shared_b_c = len(shared_b_c)

                perc_shared = round(len_shared / (len_a_only + len_b_only + len_c_only + len_shared_all + len_shared_a_b + len_shared_a_c + len_shared_b_c) * 100, 2)
                replicate_perc_shared_dict[sample] = perc_shared

                plt.figure(figsize=(10, 5))
                venn3(subsets = (len_a_only, len_b_only, len_shared_a_b, len_c_only, len_shared_a_c, len_shared_b_c, len_shared_all), set_labels = (replicate_a, replicate_b, replicate_c))
                output_pdf = Path(str(dirName) + "/" + sample + ".pdf")
                plt.savefig(output_pdf, bbox_inches='tight')
                plt.close()

            else:
                sg.Popup("Warning! More than 3 replicates are not supported for Venn diagrams:    " + sample)

            plt.close('all')

        except:
            print("Warning! No replicates found for:    " + sample)

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    samples = list(replicate_perc_shared_dict.keys())
    shared_otus = list(replicate_perc_shared_dict.values())

    fig = px.bar(x=samples, y=shared_otus, labels={"y": "Shared OTUs", "x": "Sample", "text": "Shared OTUs"}, text=shared_otus)
    fig.update_yaxes(title='Shared OTUs (%)', range=[0, 100], dtick=10, autorange=False)
    fig.update_xaxes(title='')
    fig.update_layout(width=int(width), height=int(height), template=template)
    fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1.5, opacity=opacity_value)

    ## write files
    output_pdf = Path(str(path_to_outdirs) + "/" + "Replicate_analysis" + "/" + TaXon_table_xlsx.stem + "_shared_OTUs.pdf")
    output_html = Path(str(path_to_outdirs) + "/" + "Replicate_analysis" + "/" + TaXon_table_xlsx.stem + "_shared_OTUs.html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Plots are found under:\n" + "Projects/Replicate_analysis/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write log
    from taxontabletools.create_log import ttt_log
    ttt_log("replicate analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, "nan", path_to_outdirs)



def beta_diversity(TaXon_table_xlsx, beta_w, beta_h, beta_cmap, meta_data_to_test, path_to_outdirs, template):

    import pandas as pd
    import numpy as np
    from skbio.diversity import beta_diversity
    from skbio.stats.distance import anosim
    import plotly.express as px
    from pathlib import Path
    import PySimpleGUI as sg
    import webbrowser

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
    metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()

    # check for presence absence data
    # otherwise abort and print error message
    pa_test = set([val for sublist in TaXon_table_df[TaXon_table_samples].values.tolist() for val in sublist])
    if pa_test != {1,0}:
        sg.Popup("Please use presence absence data!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == len(Meta_data_table_df['Samples'].tolist()):
        sg.Popup("The meta data is unique for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.Popup("The meta data is similar for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        data = TaXon_table_df[TaXon_table_samples].transpose().values.tolist()
        jaccard_dm = beta_diversity("jaccard", data, samples)

        anosim_results = anosim(jaccard_dm, metadata_list, permutations=999)
        anosim_r = round(anosim_results['test statistic'], 5)
        anosim_p = anosim_results['p-value']
        textbox = "Anosim (" + meta_data_to_test + ")\n" + "R = " + str(anosim_r) + "\n" + "p = " + str(anosim_p)

        matrix = jaccard_dm.data
        matrix_df = pd.DataFrame(matrix)
        matrix_df.columns = samples
        matrix_df.index = samples

        # create plot
        fig = px.imshow(matrix, x=samples,y=samples, color_continuous_scale=beta_cmap, labels=dict(color="Jaccard distance"))
        fig.update_layout(height=int(beta_h), width=int(beta_w), template=template, showlegend=True, title=textbox)

        # finish script
        output_pdf = Path(str(path_to_outdirs) + "/" + "Beta_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_jc.pdf")
        output_html = Path(str(path_to_outdirs) + "/" + "Beta_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_jc.html")
        output_xlsx = Path(str(path_to_outdirs) + "/" + "Beta_diversity" + "/" + TaXon_table_xlsx.stem + "_" + meta_data_to_test + "_jc.xlsx")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))
        matrix_df.to_excel(output_xlsx)

        ## ask to show plot
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## write to log file
        sg.Popup("Beta diversity estimate are found in", path_to_outdirs, "/Beta_diversity/", title="Finished", keep_on_top=True)
        from taxontabletools.create_log import ttt_log
        ttt_log("beta diversity", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)

    else:
        sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)



def get_available_meta_data(TaXon_table_xlsx, path_to_outdirs):

    import os
    import pandas as pd
    from pandas import DataFrame
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    meta_data_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    if os.path.exists(meta_data_xlsx):

        meta_data_xlsx = pd.ExcelFile(meta_data_xlsx)
        data = pd.read_excel(meta_data_xlsx, header=0)

        available_meta_data = data.columns.tolist()[1:]
        available_meta_data = [str(i) for i in available_meta_data]

        return available_meta_data

    else:
        return False



def create_taxon_list(TaXon_table_xlsx, taxon_list_output_file_name, create_gbif_link, calc_dist, calc_occupancy, taxon_tools_version, path_to_outdirs):

    import requests_html, json
    import re
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    from datetime import date
    import numpy as np
    from pathlib import Path
    from Bio import pairwise2
    from Bio.Seq import Seq

    # load the TaXon table
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)
    TaXon_table_df = TaXon_table_df.fillna("nan")
    samples = TaXon_table_df.columns.tolist()[10:]
    n_samples = len(samples)
    total_reads = sum([sum(reads) for reads in TaXon_table_df[samples].values.tolist()])

    # create the output files
    if taxon_list_output_file_name == '':
        taxon_list_output_file_name = "my_taxon_list"
    output_xlsx = Path(str(path_to_outdirs) + "/" + "Taxon_lists" + "/" + taxon_list_output_file_name + ".xlsx")

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Read counts and OTU numbers')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar1')],
              [sg.Text('Intraspecific distances')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar2')],
              [sg.Text('Occupancy')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar3')],
              [sg.Text('GBIF link')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar4')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar1 = window_progress_bar['progressbar1']
    progress_bar2 = window_progress_bar['progressbar2']
    progress_bar3 = window_progress_bar['progressbar3']
    progress_bar4 = window_progress_bar['progressbar4']
    ############################################################################

    # progress bar 1
    progress_update = 0
    progress_increase = 1000 / len(TaXon_table_df.values.tolist()) + 1

    # create a dict with read numbers from the TaXon table
    overall_included_OTU_set = []
    taxa_list_dict = {}
    for OTU in TaXon_table_df.values.tolist():
        species = ' '.join(OTU[1:7])
        taxonomy = OTU[1:7]
        overall_included_OTU_set.append(' '.join(taxonomy))
        n_reads = sum(OTU[10:])
        n_reads_rel = n_reads / total_reads * 100
        n_OTUs = 1
        if species not in taxa_list_dict.keys():
            taxa_list_dict[species] = taxonomy + [n_reads, n_reads_rel, n_OTUs]
        else:
            n_reads = taxa_list_dict[species][6] + n_reads
            n_reads_rel = taxa_list_dict[species][7] + n_reads_rel
            n_OTUs = taxa_list_dict[species][8] + n_OTUs
            taxa_list_dict[species] = taxonomy + [n_reads, n_reads_rel, n_OTUs]

        ############################################################################
        event, values = window_progress_bar.read(timeout=5)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar1.UpdateBar(progress_update)
        ############################################################################

    ##############################################################################
    # calculate intraspecific distances for Species with multiple OTUs

    def pw_distance(seq1, seq2):
        len1 = len(seq1)
        len2 = len(seq2)
        length = max([len1, len2])
        alignments = pairwise2.align.globalxx(seq1, seq2)
        max_score = max([alignment.score for alignment in alignments])
        diff = 100 - (max_score / length * 100)
        diff = round(diff, 2)
        return(diff)

    OTU_sequences_dict = {}
    distances_dict = {}

    # check if distance are to be calulated
    if calc_dist == True:
        for OTU in TaXon_table_df.values.tolist():
            species = OTU[6]
            if species != "nan":
                sequence = OTU[9]
                if species not in OTU_sequences_dict.keys():
                    OTU_sequences_dict[species] = [sequence]
                else:
                    OTU_sequences_dict[species] = OTU_sequences_dict[species] + [sequence]

        # progress bar 2
        progress_update = 0
        progress_increase = 1000 / len(OTU_sequences_dict.items()) + 1

        for species, sequences in OTU_sequences_dict.items():
            distances = []
            if len(sequences) > 1:
                for main_sequence in sequences:
                    for sub_sequence in sequences:
                        distances.append(pw_distance(main_sequence, sub_sequence))
                distances = list(filter((0.0).__ne__, distances))
                max_dist = max(distances)
                min_dist = min(distances)
                avg_dist = sum(distances) / len(distances)
                avg_dist = round(avg_dist, 2)
                distances_dict[species] = [max_dist, min_dist, avg_dist]

            ############################################################################
            event, values = window_progress_bar.read(timeout=5)
            if event == 'Cancel'  or event is None:
                window_progress_bar.Close()
                raise RuntimeError
            # update bar with loop value +1 so that bar eventually reaches the maximum
            progress_update += progress_increase
            progress_bar2.UpdateBar(progress_update)
            ############################################################################

    ##############################################################################
    # calculate sample occupancy

    if calc_occupancy == True:

        # progress bar 1
        progress_update = 0
        progress_increase = 1000 / len(samples) + 1

        present_OTU_list = []
        for sample in samples:
            OTUs_per_species_list = []
            # check the read abundaces for each sample
            read_abundace_list = TaXon_table_df[sample].values.tolist()
            # enumerate the read abundaces for each sample and collect all lines that have more than one read
            for i, read_abundance in enumerate(read_abundace_list):
                OTU = TaXon_table_df["IDs"][i]
                taxonomy = ' '.join(TaXon_table_df[TaXon_table_df['IDs'].str.contains(OTU + "$", regex=True)].values.tolist()[0][1:7])
                # if reads are present, collect the species name (or the specified taxonomic level) from the TaXon table
                if read_abundance != 0:
                    OTUs_per_species_list.append(taxonomy)
            # remove all nans
            OTUs_per_species_list = [x for x in OTUs_per_species_list if str(x) != 'nan']
            # make list unique
            OTUs_per_species_list = list(set(OTUs_per_species_list))
            # append to list of species for the current site
            present_OTU_list.append(OTUs_per_species_list)

            ############################################################################
            event, values = window_progress_bar.read(timeout=5)
            if event == 'Cancel'  or event is None:
                window_progress_bar.Close()
                raise RuntimeError
            # update bar with loop value +1 so that bar eventually reaches the maximum
            progress_update += progress_increase
            progress_bar3.UpdateBar(progress_update)
            ############################################################################

        # store occupancy of each species in a dict, will be accessed by position in list
        occupancy_dict = {}
        # flatten the list of present species per site
        present_OTU_list_flattened = [val for sublist in present_OTU_list for val in sublist]
        # count the number of occurences for each species and calculate the occpancy based on the number of samples
        for taxon in overall_included_OTU_set:
            count = present_OTU_list_flattened.count(taxon)
            occupancy = round(count / n_samples * 100, 2)
            if len(taxa_list_dict[taxon]) == 9:
                taxa_list_dict[taxon] = taxa_list_dict[taxon] + [occupancy]
    else:
        for taxon in overall_included_OTU_set:
            if len(taxa_list_dict[taxon]) == 9:
                taxa_list_dict[taxon] = taxa_list_dict[taxon] + [""]

    ##############################################################################

    def gbif_requester(species_name):

        ## create an html session
        with requests_html.HTMLSession() as session:
            ## generate html request name
            request_name = '%20'.join(species_name.split(' '))

            ## request that name
            r = session.get('https://api.gbif.org/v1/species?name=%s&limit=1' % request_name)

            ## parse json
            res = json.loads(r.text)

            ## return the link
            if res['results']:
                return 'https://www.gbif.org/species/%s' % res['results'][0]['key']
            else:
                return False

    gbif_df_list = []

    # progress bar 3
    progress_update = 0
    progress_increase = 1000 / len(taxa_list_dict.items()) + 1

    if create_gbif_link == True:

        for taxon, values in taxa_list_dict.items():
            species = values[-5]
            if create_gbif_link == True:
                if species != 'nan':
                    gbif_link = gbif_requester(species)
                    if gbif_link != False:
                        species_entry = values + [gbif_link]
                        gbif_df_list.append(species_entry)
                    else:
                        species_entry = values + [""]
                        gbif_df_list.append(species_entry)
                else:
                    species_entry = values + [""]
                    gbif_df_list.append(species_entry)
            else:
                species_entry = values + [""]
                gbif_df_list.append(species_entry)

            ############################################################################
            event, values = window_progress_bar.read(timeout=10)
            if event == 'Cancel'  or event is None:
                window_progress_bar.Close()
                raise RuntimeError
            # update bar with loop value +1 so that bar eventually reaches the maximum
            progress_update += progress_increase
            progress_bar4.UpdateBar(progress_update)
            ############################################################################

    else:
        for taxon, values in taxa_list_dict.items():
            species_entry = values + [""]
            gbif_df_list.append(species_entry)

    window_progress_bar.Close()

    # now append the distances for each species that has multiple OTUs
    # create the final list for the output dataframes
    df_out_list = []
    for entry in gbif_df_list:
        species = entry[5]
        if species in distances_dict.keys():
            df_out_list.append(entry + distances_dict[species])
        else:
            df_out_list.append(entry + ["", "", ""])

    df_out  = pd.DataFrame(df_out_list, columns=TaXon_table_df.columns.tolist()[1:7] + ["reads", "reads (%)", "OTUs", "occupancy (%)", "gbif", "dist max (%)", "dist min (%)", "dist avg (%)"])
    df_out = df_out[TaXon_table_df.columns.tolist()[1:7] + ["reads", "reads (%)", "OTUs", "occupancy (%)", "dist max (%)", "dist min (%)", "dist avg (%)", "gbif"]]
    df_out.to_excel(output_xlsx, sheet_name = 'Taxa', index=False)

    closing_text = "Taxon list is found under:\n" + '/'.join(str(output_xlsx).split("/")[-4:])
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("taxon list", "analysis", TaXon_table_xlsx.name, output_xlsx.name, "", path_to_outdirs)



def PCoA_analysis(TaXon_table_xlsx, meta_data_to_test, width, height, pcoa_s, draw_mesh, path_to_outdirs, template):
    import pandas as pd
    import numpy as np
    from skbio.diversity import beta_diversity
    from skbio.stats.ordination import pcoa
    from skbio.stats.distance import anosim
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    import plotly.express as px
    from pathlib import Path
    import PySimpleGUI as sg
    import os, webbrowser

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    # check for presence absence data
    # otherwise abort and print error message
    pa_test = set([val for sublist in TaXon_table_df[TaXon_table_samples].values.tolist() for val in sublist])
    if pa_test != {1,0}:
        sg.Popup("Please use presence absence data!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == len(Meta_data_table_df['Samples'].tolist()):
        sg.Popup("The meta data is unique for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.Popup("The meta data is similar for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples
        data = TaXon_table_df[TaXon_table_samples].transpose().values.tolist()
        jc_dm = beta_diversity("jaccard", data, samples)
        ordination_result = pcoa(jc_dm)
        metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()

        anosim_results = anosim(jc_dm, metadata_list, permutations=999)
        anosim_r = round(anosim_results['test statistic'], 5)
        anosim_p = anosim_results['p-value']
        textbox = meta_data_to_test + "\nAnosim " + "R = " + str(anosim_r) + " " + "p = " + str(anosim_p)

        #######################################################################################
        # create window to ask for PCoA axis to test
        def slices(list, slice):
            for i in range(0, len(list), slice):
                yield list[i : i + slice]

        # collect the PCoA proportion explained values
        proportion_explained_list = []
        for i, pcoa_axis in enumerate(ordination_result.proportion_explained):
            if round(pcoa_axis* 100, 2) >= 1:
                proportion_explained_list.append("PC" + str(i+1) + " (" + str(round(pcoa_axis* 100, 2)) + " %)")

        pcoa_axis_checkboxes = list(slices([sg.Checkbox(name, key=name, size=(15,1)) for name in proportion_explained_list], 10))

        pcoa_window_layout = [
                    [sg.Text('Check two axis to be displayed')],
                    [sg.Frame(layout = pcoa_axis_checkboxes, title = '')],
                    [sg.Text('Only axes >= 1 % explained variance are shown')],
                    [sg.Text('')],
                    [sg.Button('Plot', key='Plot'), sg.Text(''), sg.Button('Plot matrix')],
                    [sg.Button('Back')],
                    ]

        pcoa_window = sg.Window('PCoA axis', pcoa_window_layout, keep_on_top=True)

        while True:
            event, values = pcoa_window.read()

            if event is None or event == 'Back':
                break

            if event == 'Plot':

                ## create a subfolder for better sorting and overview
                dirName = Path(str(path_to_outdirs) + "/" + "PCoA_plots" + "/" + TaXon_table_xlsx.stem + "/")
                if not os.path.exists(dirName):
                    os.mkdir(dirName)

                # collect the pcoa axis values
                axis_to_plot = [key for key,value in values.items() if value == True and "PC" in key]
                # pass on only if two pcoa axes were checked
                if len(axis_to_plot) == 2:
                    cat1 = axis_to_plot[1].split()[0]
                    cat2 = axis_to_plot[0].split()[0]

                    df_pcoa = ordination_result.samples[[cat1, cat2]]
                    df_pcoa.insert(2, "Metadata", Meta_data_table_df[meta_data_to_test].values.tolist(), True)
                    df_pcoa.insert(3, "Samples", Meta_data_table_df["Samples"].values.tolist(), True)

                    fig = px.scatter(df_pcoa, x=cat1, y=cat2, color="Metadata", text="Samples", title=textbox)
                    fig.update_traces(marker_size=int(pcoa_s), mode="markers")
                    fig.update_layout(height=int(height), width=int(width), template=template, showlegend=True)
                    fig.update_xaxes(title=axis_to_plot[1])
                    fig.update_yaxes(title=axis_to_plot[0])

                    ## define output files
                    output_pdf = Path(str(dirName) + "/" + meta_data_to_test + ".pdf")
                    output_html = Path(str(dirName) + "/" + meta_data_to_test + ".html")
                    output_xlsx = Path(str(dirName) + "/" + meta_data_to_test + ".xlsx")

                    ## write files
                    fig.write_image(str(output_pdf))
                    fig.write_html(str(output_html))
                    ordination_result.samples[[cat1, cat2]].to_excel(output_xlsx)

                    ## ask to show file
                    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
                    if answer == "Yes":
                        webbrowser.open('file://' + str(output_html))

                    ## print closing text
                    closing_text = "\n" + "PCoA plots are found in: " + str(path_to_outdirs) + "/PCoA_plots/"
                    sg.Popup(closing_text, title="Finished", keep_on_top=True)

                    ## write to log
                    from taxontabletools.create_log import ttt_log
                    ttt_log("pcoa analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)
                    break

                elif len(axis_to_plot) == 3:
                    cat1 = axis_to_plot[0].split()[0]
                    cat2 = axis_to_plot[1].split()[0]
                    cat3 = axis_to_plot[2].split()[0]

                    df_pcoa = ordination_result.samples[[cat1, cat2, cat3]]
                    df_pcoa.insert(3, "Metadata", Meta_data_table_df[meta_data_to_test].values.tolist(), True)
                    df_pcoa.insert(4, "Samples", Meta_data_table_df["Samples"].values.tolist(), True)

                    from itertools import combinations

                    ## check if lines are to be drawn between the dots
                    if draw_mesh == True:
                        combinations_list =[]
                        for metadata in df_pcoa["Metadata"]:
                            ## collect all entries for the respective metadata
                            arr = df_pcoa.loc[df_pcoa['Metadata'] == metadata][[cat1, cat2, cat3, "Metadata", "Samples"]].to_numpy()
                            ## create a df for all possible combinations using itertools combinations
                            for entry in list(combinations(arr, 2)):
                                combinations_list.append(list(entry[0]))
                                combinations_list.append(list(entry[1]))
                        ## create a dataframe to draw the plot from
                        df = pd.DataFrame(combinations_list)
                        df.columns = [cat1, cat2, cat3, "Metadata", "Samples"]
                        ## draw the plot
                        fig = px.scatter_3d(df, x=cat1, y=cat2, z=cat3, color="Metadata", text="Samples", title=textbox)
                        fig.update_traces(marker_size=int(pcoa_s), mode="markers+lines", line=dict(width=0.5))
                        fig.update_layout(height=int(height), width=int(width), template=template, showlegend=True)
                        fig.update_layout(scene = dict(xaxis_title=axis_to_plot[0],yaxis_title=axis_to_plot[1],zaxis_title=axis_to_plot[2]))
                    else:
                        fig = px.scatter_3d(df_pcoa, x=cat1, y=cat2, z=cat3, color="Metadata", text="Samples")
                        fig.update_traces(marker_size=int(pcoa_s), mode="markers")
                        fig.update_layout(height=int(height), width=int(width), template=template, showlegend=True, title=textbox)
                        fig.update_layout(scene = dict(xaxis_title=axis_to_plot[0],yaxis_title=axis_to_plot[1],zaxis_title=axis_to_plot[2]))

                    ## define output files
                    output_pdf = Path(str(dirName) + "/" + meta_data_to_test + "_3d.pdf")
                    output_html = Path(str(dirName) + "/" + meta_data_to_test + "_3d.html")
                    output_xlsx = Path(str(dirName) + "/" + meta_data_to_test + "_3d.xlsx")

                    ## write output files
                    fig.write_image(str(output_pdf))
                    fig.write_html(str(output_html))
                    ordination_result.samples[[cat1, cat2]].to_excel(output_xlsx)

                    ## ask to show file
                    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
                    if answer == "Yes":
                        webbrowser.open('file://' + str(output_html))

                    ## print closing text
                    closing_text = "PCoA plots are found in: " + str(path_to_outdirs) + "/PCoA_plots/"
                    sg.Popup(closing_text, title="Finished", keep_on_top=True)

                    ## write log file
                    from taxontabletools.create_log import ttt_log
                    ttt_log("pcoa analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)
                    break

                else:
                    sg.Popup("Please choose not more than 3 PCoA axes", title="Error", keep_on_top=True)

            if event == 'Plot matrix':
                if len(proportion_explained_list) >= 4:

                    ## create a subfolder for better sorting and overview
                    dirName = Path(str(path_to_outdirs) + "/" + "PCoA_plots" + "/" + TaXon_table_xlsx.stem + "/")
                    if not os.path.exists(dirName):
                        os.mkdir(dirName)

                    df_pcoa = ordination_result.samples[["PC1", "PC2", "PC3", "PC4"]]
                    df_pcoa.insert(4, "Metadata", Meta_data_table_df[meta_data_to_test].values.tolist(), True)
                    df_pcoa.insert(5, "Sample", Meta_data_table_df["Samples"].values.tolist(), True)

                    fig = make_subplots(rows=4, cols=4)
                    ########### 1 ###########
                    fig.add_trace(go.Scatter(),row=1, col=1)
                    fig.update_layout(template=template)
                    text = "PC1 (" + str(round(ordination_result.proportion_explained["PC1"]* 100, 2)) + " %)"
                    fig.add_annotation(text=text, showarrow=False)
                    fig.update_xaxes(showticklabels=False, showgrid=False)
                    fig.update_yaxes(showticklabels=False, showgrid=False)
                    ########### 2 ###########
                    df = df_pcoa[["PC1", "PC2", "Metadata", "Sample"]]
                    for metadata in set(metadata_list):
                        df_metadata = df[df['Metadata'] == metadata]
                        #fig = px.scatter(df_pcoa, x="PC1", y="PC2", , )
                        fig.add_trace(go.Scatter(   x=df_metadata["PC1"].values.tolist(),
                                                    y=df_metadata["PC2"].values.tolist(),
                                                    mode='markers',
                                                    name=metadata,
                                                    text=df_metadata["Sample"].values.tolist()),row=1, col=2)
                    ########### 3 ###########
                    df = df_pcoa[["PC1", "PC3", "Metadata", "Sample"]]
                    for metadata in set(metadata_list):
                        df_metadata = df[df['Metadata'] == metadata]
                        #fig = px.scatter(df_pcoa, x="PC1", y="PC2", , )
                        fig.add_trace(go.Scatter(   x=df_metadata["PC1"].values.tolist(),
                                                    y=df_metadata["PC3"].values.tolist(),
                                                    mode='markers',
                                                    name=metadata,
                                                    showlegend=False,
                                                    text=df_metadata["Sample"].values.tolist()),row=1, col=3)
                    ########### 4 ###########
                    df = df_pcoa[["PC1", "PC4", "Metadata", "Sample"]]
                    for metadata in set(metadata_list):
                        df_metadata = df[df['Metadata'] == metadata]
                        fig.add_trace(go.Scatter(   x=df_metadata["PC1"].values.tolist(),
                                                    y=df_metadata["PC4"].values.tolist(),
                                                    mode='markers',
                                                    name=metadata,
                                                    showlegend=False,
                                                    text=df_metadata["Sample"].values.tolist()),row=1, col=4)
                        fig.update_traces(marker_size=int(pcoa_s), mode="markers")
                        fig.update_xaxes(showgrid=False,row=1, col=4)
                        fig.update_yaxes(showgrid=False,row=1, col=4)
                    ########### 5 ###########
                    fig.add_trace(go.Scatter(),row=2, col=2)
                    fig.update_layout(template=template)
                    text = "PC2 (" + str(round(ordination_result.proportion_explained["PC2"]* 100, 2)) + " %)"
                    fig.add_annotation(text=text, showarrow=False, row=2, col=2)
                    ########### 6 ###########
                    df = df_pcoa[["PC2", "PC3", "Metadata", "Sample"]]
                    for metadata in set(metadata_list):
                        df_metadata = df[df['Metadata'] == metadata]
                        #fig = px.scatter(df_pcoa, x="PC1", y="PC2", , )
                        fig.add_trace(go.Scatter(   x=df_metadata["PC2"].values.tolist(),
                                                    y=df_metadata["PC3"].values.tolist(),
                                                    mode='markers',
                                                    name=metadata,
                                                    showlegend=False,
                                                    text=df_metadata["Sample"].values.tolist()),row=2, col=3)
                    ########### 7 ###########
                    df = df_pcoa[["PC2", "PC4", "Metadata", "Sample"]]
                    for metadata in set(metadata_list):
                        df_metadata = df[df['Metadata'] == metadata]
                        fig.add_trace(go.Scatter(   x=df_metadata["PC2"].values.tolist(),
                                                    y=df_metadata["PC4"].values.tolist(),
                                                    mode='markers',
                                                    name=metadata,
                                                    showlegend=False,
                                                    text=df_metadata["Sample"].values.tolist()),row=2, col=4)
                    ########### 8 ###########
                    fig.add_trace(go.Scatter(),row=3, col=3)
                    fig.update_layout(template=template)
                    text = "PC3 (" + str(round(ordination_result.proportion_explained["PC3"]* 100, 2)) + " %)"
                    fig.add_annotation(text=text, showarrow=False, row=3, col=3)
                    ########### 9 ###########
                    df = df_pcoa[["PC3", "PC4", "Metadata", "Sample"]]
                    for metadata in set(metadata_list):
                        df_metadata = df[df['Metadata'] == metadata]
                        #fig = px.scatter(df_pcoa, x="PC1", y="PC2", , )
                        fig.add_trace(go.Scatter(   x=df_metadata["PC3"].values.tolist(),
                                                    y=df_metadata["PC4"].values.tolist(),
                                                    mode='markers',
                                                    name=metadata,
                                                    showlegend=False,
                                                    text=df_metadata["Sample"].values.tolist()),row=3, col=4)
                    ########### 5 ###########
                    fig.add_trace(go.Scatter(),row=4, col=4)
                    fig.update_layout(template=template)
                    text = "PC4 (" + str(round(ordination_result.proportion_explained["PC4"]* 100, 2)) + " %)"
                    fig.add_annotation(text=text, showarrow=False, row=4, col=4)

                    ######################
                    fig.update_xaxes(showline=True, mirror=True, linewidth=1, linecolor='black')
                    fig.update_yaxes(showline=True, mirror=True, linewidth=1, linecolor='black')
                    fig.update_traces(marker_size=int(pcoa_s), mode="markers")
                    # finish plot matrix
                    fig.update_layout(height=1000, width=1000, title_text=textbox)

                    ## define output files
                    output_pdf = Path(str(dirName) + "/" + meta_data_to_test + "_matrix.pdf")
                    output_html = Path(str(dirName) + "/" + meta_data_to_test + "_matrix.html")

                    ## write output files
                    fig.write_image(str(output_pdf))
                    fig.write_html(str(output_html))

                    ## ask to show file
                    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
                    if answer == "Yes":
                        webbrowser.open('file://' + str(output_html))

                    ## print closing text
                    closing_text = "\n" + "PCoA plots are found in: " + str(path_to_outdirs) + "/PCoA_plots/"
                    sg.Popup(closing_text, title="Finished", keep_on_top=True)

                    ## write to log file
                    from taxontabletools.create_log import ttt_log
                    ttt_log("pcoa analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)
                    break
                else:
                    sg.Popup("There must be at least 4 PCoA axis available to plot the matrix!")

        pcoa_window.close()

    else:
        sg.PopupError("The sample of both the TaXon table and the metadata table have to match!")



def gbif_parent_check(phylum_name, taxon_name, taxonomy_check):

    import requests_html, json
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path

    #taxon_name = "Radix ampla" taxon_level = "species"
    ## create an html session
    with requests_html.HTMLSession() as session:
        ## generate html request name
        request_name = '%20'.join(taxon_name.split(' '))
        ## request that name
        r = session.get("https://api.gbif.org/v1/species/match?verbose=true&name=" + request_name + "&limit=1")
        ## parse json
        res = json.loads(r.text)

        gbif_result = []

        if 'note' in res.keys():
            if "Multiple equal matches" in res['note']:
                for match in res["alternatives"]:
                    if phylum_name == match["phylum"]:
                        try:
                            for taxon_level in taxonomy_check:
                                    gbif_result.append(match[taxon_level.lower()])
                            return gbif_result
                            break
                        except:
                            gbif_result.append("")
                            return gbif_result
                            break
            else:
                for taxon_level in taxonomy_check:
                    try:
                        gbif_result.append(res[taxon_level.lower()])
                    except:
                        try:
                            gbif_result.append(res["alternatives"][0][taxon_level.lower()])
                        except:
                            gbif_result.append("")
                return gbif_result
        else:
            gbif_result.append("")
            return gbif_result

def gbif_check_taxonomy(TaXon_table_xlsx, path_to_outdirs):

    import requests_html, json
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    taxon_levels = ["Phylum","Class","Order","Family","Genus","Species"]
    OTUs_list = TaXon_table_df["IDs"].values.tolist()

    taxonomy_check_dict = {}

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(OTUs_list) + 1
    ############################################################################

    for OTU in TaXon_table_df[["Phylum","Class","Order","Family","Genus","Species"]].fillna("").values.tolist():

        for i, taxonomy in enumerate(OTU):
            if taxonomy == "":
                phylum_name = OTU[0]
                taxon_name = OTU[i-1]
                taxonomy_check = taxon_levels[0: i]
                result = gbif_parent_check(phylum_name, taxon_name, taxonomy_check)
                query = OTU[0: i]
                if query != result:
                    if len(query) != 6:
                        add = 6 - len(query)
                        query = query + [''] * add
                    if len(result) != 6:
                        add = 6 - len(result)
                        result = result + [''] * add
                    query = ",".join(query)
                    taxonomy_check_dict[query] = result
                break

            elif i == 5:
                phylum_name = OTU[0]
                taxon_name = OTU[5]
                taxonomy_check = taxon_levels
                result = gbif_parent_check(phylum_name, taxon_name, taxonomy_check)
                if OTU != result:
                    query = ",".join(OTU)
                    taxonomy_check_dict[query] = result

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    TaXon_table_list = []

    for OTU in TaXon_table_df.fillna("").values.tolist():
        taxonomy = OTU[1:7]
        search_key = ','.join(taxonomy)
        if (search_key in taxonomy_check_dict.keys() and taxonomy_check_dict[search_key] != ['']*6):
            replacement_taxonomy = taxonomy_check_dict[search_key]
            replacement_OTU = [OTU[0]] + replacement_taxonomy + OTU[7:]
            TaXon_table_list.append(replacement_OTU)
        else:
            TaXon_table_list.append(OTU)

    file_name = TaXon_table_xlsx.stem
    output_name = Path(str(path_to_outdirs) + "/TaXon_tables/" + file_name + "_gbif" + ".xlsx")
    df_new = pd.DataFrame(TaXon_table_list, columns=(TaXon_table_df.columns.values.tolist()))
    df_new.to_excel(output_name, sheet_name = 'TaXon table', index=False)

    change_log_list = []
    for key, value in taxonomy_check_dict.items():
        change_log_list.append(["Input:"] + key.split(","))
        change_log_list.append(["Gbif:"] + value)

    change_log_df = pd.DataFrame(change_log_list, columns=(["Change"] + taxon_levels))
    change_log_name = Path(str(path_to_outdirs) + "/GBIF/" + file_name + "_gbif_log" + ".xlsx")
    change_log_df = pd.DataFrame(change_log_list, columns=(["Change"] + taxon_levels))
    change_log_df.to_excel(change_log_name, sheet_name = 'TaXon table', index=False)

    closing_text = "Taxon table is found under:\n" + '/'.join(str(output_name).split("/")[-4:]) + "\n\n" + "Log file is found under:\n" + '/'.join(str(change_log_name).split("/")[-4:])
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("gbif check", "processing", TaXon_table_xlsx.name, output_name.name, "nan", path_to_outdirs)



def replicate_consistency_filter(TaXon_table_xlsx, suffix_list, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)

    sample_names = TaXon_table_df.columns[10:].tolist()
    OTUs = TaXon_table_df["IDs"].values.tolist()

    derep_sample_names_dict =  {}
    unique_sample_names_list = []
    replicates_dict = {}

    for sample in sample_names:
        sample_name = sample.split("_")[0:-1]
        unique_sample_names_list.append("_".join(sample_name))

    unique_sample_names_set = sorted(set(unique_sample_names_list))

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(unique_sample_names_set) + 1
    ############################################################################

    for sample in unique_sample_names_set:

        for i, suffix in enumerate(suffix_list):
            replicates_dict["rep_" + str(i)] = sample + "_" + suffix_list[i]

        replicate_names_list = list(replicates_dict.values())

        try:
            new_df = TaXon_table_df[replicate_names_list]
            header = new_df.columns.tolist()
            processed_reads = []

            for n_reads in new_df.values.tolist():
                if 0 in n_reads:
                    if len(set(n_reads)) > 1:
                        n_reads = len(n_reads) * [0]
                processed_reads.append(n_reads)

            df_out = pd.DataFrame(processed_reads)
            df_out.columns = header
            TaXon_table_df = TaXon_table_df.drop(replicate_names_list, axis=1)
            TaXon_table_df[replicate_names_list] = df_out

        except:
            print("Warning! No replicates found for:    " + sample)

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            print('Cancel')
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    # filter for 0 hit OTUs (can happen after consistency filtering)
    columns = TaXon_table_df.columns.tolist()
    TaXon_table_list = TaXon_table_df.values.tolist()
    TaXon_table_list_final = []
    for entry in TaXon_table_list:
        if sum(entry[10:]) != 0:
            TaXon_table_list_final.append(entry)
        else:
            print("Dropped:", entry[0], "(0 reads)")

    TaXon_table_df = pd.DataFrame(TaXon_table_list_final, columns=columns)

    taxon_tables_directory = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + TaXon_table_xlsx.stem)
    taxon_tables_cons_xlsx = Path(str(taxon_tables_directory) + "_cons.xlsx")

    TaXon_table_df.to_excel(taxon_tables_cons_xlsx, sheet_name='TaXon table', index=False)

    closing_text = "Taxon table is found under:\n" + '/'.join(str(taxon_tables_cons_xlsx).split("/")[-4:])
    print(closing_text)
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("replicate consistency", "processing", TaXon_table_xlsx.name, taxon_tables_cons_xlsx.name, "nan", path_to_outdirs)



def gbif_occurrence(TaXon_table_xlsx, width, height, continents_to_check, template, theme, path_to_outdirs):

    import requests_html, json
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path
    import plotly.graph_objects as go
    import os, webbrowser

    ## dictionary with all country codes of the Earth
    country_codes_dict = {'Andorra': ['AD', 'Europe'],
 'United Arab Emirates': ['AE', 'Asia'],
 'Afghanistan': ['AF', 'Asia'],
 'Antigua and Barbuda': ['AG', 'North America'],
 'Anguilla': ['AI', 'North America'],
 'Albania': ['AL', 'Europe'],
 'Armenia': ['AM', 'Asia'],
 'Angola': ['AO', 'Africa'],
 'Antarctica': ['AQ', 'Antarctica'],
 'Argentina': ['AR', 'South America'],
 'American Samoa': ['AS', 'Oceania'],
 'Austria': ['AT', 'Europe'],
 'Australia': ['AU', 'Oceania'],
 'Aruba': ['AW', 'North America'],
 'Åland Islands': ['AX', 'Europe'],
 'Azerbaijan': ['AZ', 'Asia'],
 'Bosnia and Herzegovina': ['BA', 'Europe'],
 'Barbados': ['BB', 'North America'],
 'Bangladesh': ['BD', 'Asia'],
 'Belgium': ['BE', 'Europe'],
 'Burkina Faso': ['BF', 'Africa'],
 'Bulgaria': ['BG', 'Europe'],
 'Bahrain': ['BH', 'Asia'],
 'Burundi': ['BI', 'Africa'],
 'Benin': ['BJ', 'Africa'],
 'Saint Barthélemy': ['BL', 'North America'],
 'Bermuda': ['BM', 'North America'],
 'Brunei Darussalam': ['BN', 'Asia'],
 'Bolivia': ['BO', 'South America'],
 'Bonaire, Sint Eustatius and Saba': ['BQ', 'North America'],
 'Brazil': ['BR', 'South America'],
 'Bahamas': ['BS', 'North America'],
 'Bhutan': ['BT', 'Asia'],
 'Bouvet Island': ['BV', 'Antarctica'],
 'Botswana': ['BW', 'Africa'],
 'Belarus': ['BY', 'Europe'],
 'Belize': ['BZ', 'North America'],
 'Canada': ['CA', 'North America'],
 'Cocos (Keeling) Islands': ['CC', 'Asia'],
 'Congo (Democratic Republic)': ['CD', 'Africa'],
 'Central African Republic': ['CF', 'Africa'],
 'Congo': ['CG', 'Africa'],
 'Switzerland': ['CH', 'Europe'],
 "Côte d'Ivoire": ['CI', 'Africa'],
 'Cook Islands': ['CK', 'Oceania'],
 'Chile': ['CL', 'South America'],
 'Cameroon': ['CM', 'Africa'],
 'China': ['CN', 'Asia'],
 'Colombia': ['CO', 'South America'],
 'Costa Rica': ['CR', 'North America'],
 'Cuba': ['CU', 'North America'],
 'Cabo Verde': ['CV', 'Africa'],
 'Curaçao': ['CW', 'North America'],
 'Christmas Island': ['CX', 'Asia'],
 'Cyprus': ['CY', 'Asia'],
 'Czechia': ['CZ', 'Europe'],
 'Germany': ['DE', 'Europe'],
 'Djibouti': ['DJ', 'Africa'],
 'Denmark': ['DK', 'Europe'],
 'Dominica': ['DM', 'North America'],
 'Dominican Republic': ['DO', 'North America'],
 'Algeria': ['DZ', 'Africa'],
 'Ecuador': ['EC', 'South America'],
 'Estonia': ['EE', 'Europe'],
 'Egypt': ['EG', 'Africa'],
 'Western Sahara': ['EH', 'Africa'],
 'Eritrea': ['ER', 'Africa'],
 'Spain': ['ES', 'Europe'],
 'Ethiopia': ['ET', 'Africa'],
 'Finland': ['FI', 'Europe'],
 'Fiji': ['FJ', 'Oceania'],
 'Falkland Islands': ['FK', 'South America'],
 'Micronesia': ['FM', 'Oceania'],
 'Faroe Islands': ['FO', 'Europe'],
 'France': ['FR', 'Europe'],
 'Gabon': ['GA', 'Africa'],
 'United Kingdom': ['GB',
  'Europe'],
 'Grenada': ['GD', 'North America'],
 'Georgia': ['GE', 'Asia'],
 'French Guiana': ['GF', 'South America'],
 'Guernsey': ['GG', 'Europe'],
 'Ghana': ['GH', 'Africa'],
 'Gibraltar': ['GI', 'Europe'],
 'Greenland': ['GL', 'North America'],
 'Gambia': ['GM', 'Africa'],
 'Guinea': ['GN', 'Africa'],
 'Guadeloupe': ['GP', 'North America'],
 'Equatorial Guinea': ['GQ', 'Africa'],
 'Greece': ['GR', 'Europe'],
 'South Georgia and the South Sandwich Islands': ['GS', 'Antarctica'],
 'Guatemala': ['GT', 'North America'],
 'Guam': ['GU', 'Oceania'],
 'Guinea-Bissau': ['GW', 'Africa'],
 'Guyana': ['GY', 'South America'],
 'Hong Kong': ['HK', 'Asia'],
 'Heard Island and McDonald Islands': ['HM', 'Antarctica'],
 'Honduras': ['HN', 'North America'],
 'Croatia': ['HR', 'Europe'],
 'Haiti': ['HT', 'North America'],
 'Hungary': ['HU', 'Europe'],
 'Indonesia': ['ID', 'Asia'],
 'Ireland': ['IE', 'Europe'],
 'Israel': ['IL', 'Asia'],
 'Isle of Man': ['IM', 'Europe'],
 'India': ['IN', 'Asia'],
 'British Indian Ocean Territory': ['IO', 'Asia'],
 'Iraq': ['IQ', 'Asia'],
 'Iran': ['IR', 'Asia'],
 'Iceland': ['IS', 'Europe'],
 'Italy': ['IT', 'Europe'],
 'Jersey': ['JE', 'Europe'],
 'Jamaica': ['JM', 'North America'],
 'Jordan': ['JO', 'Asia'],
 'Japan': ['JP', 'Asia'],
 'Kenya': ['KE', 'Africa'],
 'Kyrgyzstan': ['KG', 'Asia'],
 'Cambodia': ['KH', 'Asia'],
 'Kiribati': ['KI', 'Oceania'],
 'Comoros': ['KM', 'Africa'],
 'Saint Kitts and Nevis': ['KN', 'North America'],
 "Korea (Democratic People's Republic)": ['KP', 'Asia'],
 'Korea (Republic)': ['KR', 'Asia'],
 'Kuwait': ['KW', 'Asia'],
 'Cayman Islands': ['KY', 'North America'],
 'Kazakhstan': ['KZ', 'Asia'],
 "Lao People's Democratic Republic": ['LA', 'Asia'],
 'Lebanon': ['LB', 'Asia'],
 'Saint Lucia': ['LC', 'North America'],
 'Liechtenstein': ['LI', 'Europe'],
 'Sri Lanka': ['LK', 'Asia'],
 'Liberia': ['LR', 'Africa'],
 'Lesotho': ['LS', 'Africa'],
 'Lithuania': ['LT', 'Europe'],
 'Luxembourg': ['LU', 'Europe'],
 'Latvia': ['LV', 'Europe'],
 'Libya': ['LY', 'Africa'],
 'Morocco': ['MA', 'Africa'],
 'Monaco': ['MC', 'Europe'],
 'Moldova (the Republic of)': ['MD', 'Europe'],
 'Montenegro': ['ME', 'Europe'],
 'Saint Martin (French part)': ['MF', 'North America'],
 'Madagascar': ['MG', 'Africa'],
 'Marshall Islands': ['MH', 'Oceania'],
 'Republic of North Macedonia': ['MK', 'Europe'],
 'Mali': ['ML', 'Africa'],
 'Myanmar': ['MM', 'Asia'],
 'Mongolia': ['MN', 'Asia'],
 'Macao': ['MO', 'Asia'],
 'Northern Mariana Islands': ['MP', 'Oceania'],
 'Martinique': ['MQ', 'North America'],
 'Mauritania': ['MR', 'Africa'],
 'Montserrat': ['MS', 'North America'],
 'Malta': ['MT', 'Europe'],
 'Mauritius': ['MU', 'Africa'],
 'Maldives': ['MV', 'Asia'],
 'Malawi': ['MW', 'Africa'],
 'Mexico': ['MX', 'North America'],
 'Malaysia': ['MY', 'Asia'],
 'Mozambique': ['MZ', 'Africa'],
 'Namibia': ['NA', 'Africa'],
 'New Caledonia': ['NC', 'Oceania'],
 'Niger': ['NE', 'Africa'],
 'Norfolk Island': ['NF', 'Oceania'],
 'Nigeria': ['NG', 'Africa'],
 'Nicaragua': ['NI', 'North America'],
 'Netherlands': ['NL', 'Europe'],
 'Norway': ['NO', 'Europe'],
 'Nepal': ['NP', 'Asia'],
 'Nauru': ['NR', 'Oceania'],
 'Niue': ['NU', 'Oceania'],
 'New Zealand': ['NZ', 'Oceania'],
 'Oman': ['OM', 'Asia'],
 'Panama': ['PA', 'North America'],
 'Peru': ['PE', 'South America'],
 'French Polynesia': ['PF', 'Oceania'],
 'Papua New Guinea': ['PG', 'Oceania'],
 'Philippines': ['PH', 'Asia'],
 'Pakistan': ['PK', 'Asia'],
 'Poland': ['PL', 'Europe'],
 'Saint Pierre and Miquelon': ['PM', 'North America'],
 'Pitcairn': ['PN', 'Oceania'],
 'Puerto Rico': ['PR', 'North America'],
 'Palestine, State of': ['PS', 'Asia'],
 'Portugal': ['PT', 'Europe'],
 'Palau': ['PW', 'Oceania'],
 'Paraguay': ['PY', 'South America'],
 'Qatar': ['QA', 'Asia'],
 'Réunion': ['RE', 'Africa'],
 'Romania': ['RO', 'Europe'],
 'Serbia': ['RS', 'Europe'],
 'Russian Federation': ['RU', 'Europe'],
 'Rwanda': ['RW', 'Africa'],
 'Saudi Arabia': ['SA', 'Asia'],
 'Solomon Islands': ['SB', 'Oceania'],
 'Seychelles': ['SC', 'Africa'],
 'Sudan': ['SD', 'Africa'],
 'Sweden': ['SE', 'Europe'],
 'Singapore': ['SG', 'Asia'],
 'Saint Helena, Ascension and Tristan da Cunha': ['SH', 'Africa'],
 'Slovenia': ['SI', 'Europe'],
 'Svalbard and Jan Mayen': ['SJ', 'Europe'],
 'Slovakia': ['SK', 'Europe'],
 'Sierra Leone': ['SL', 'Africa'],
 'San Marino': ['SM', 'Europe'],
 'Senegal': ['SN', 'Africa'],
 'Somalia': ['SO', 'Africa'],
 'Suriname': ['SR', 'South America'],
 'South Sudan': ['SS', 'Africa'],
 'Sao Tome and Principe': ['ST', 'Africa'],
 'El Salvador': ['SV', 'North America'],
 'Syrian Arab Republic': ['SY', 'Asia'],
 'Eswatini': ['SZ', 'Africa'],
 'Turks and Caicos Islands': ['TC', 'North America'],
 'Chad': ['TD', 'Africa'],
 'French Southern Territories': ['TF', 'Antarctica'],
 'Togo': ['TG', 'Africa'],
 'Thailand': ['TH', 'Asia'],
 'Tajikistan': ['TJ', 'Asia'],
 'Tokelau': ['TK', 'Oceania'],
 'Timor-Leste': ['TL', 'Asia'],
 'Turkmenistan': ['TM', 'Asia'],
 'Tunisia': ['TN', 'Africa'],
 'Tonga': ['TO', 'Oceania'],
 'Turkey': ['TR', 'Europe'],
 'Trinidad and Tobago': ['TT', 'North America'],
 'Tuvalu': ['TV', 'Oceania'],
 'Taiwan': ['TW', 'Asia'],
 'Tanzania': ['TZ', 'Africa'],
 'Ukraine': ['UA', 'Europe'],
 'Uganda': ['UG', 'Africa'],
 'United States Minor Outlying Islands': ['UM', 'Oceania'],
 'United States of America': ['US', 'North America'],
 'Uruguay': ['UY', 'South America'],
 'Uzbekistan': ['UZ', 'Asia'],
 'Holy See': ['VA', 'Europe'],
 'Saint Vincent and the Grenadines': ['VC', 'North America'],
 'Venezuela (Bolivarian Republic of)': ['VE', 'South America'],
 'Virgin Islands (British)': ['VG', 'North America'],
 'Virgin Islands (U.S.)': ['VI', 'North America'],
 'Viet Nam': ['VN', 'Asia'],
 'Vanuatu': ['VU', 'Oceania'],
 'Wallis and Futuna': ['WF', 'Oceania'],
 'Samoa': ['WS', 'Oceania'],
 'Yemen': ['YE', 'Asia'],
 'Mayotte': ['YT', 'Africa'],
 'South Africa': ['ZA', 'Africa'],
 'Zambia': ['ZM', 'Africa'],
 'Zimbabwe': ['ZW', 'Africa']}

    ## load Taxon table
    #TaXon_table_xlsx = "/Users/tillmacher/Desktop/Projects/TTT_Projects/Projects/Sicliy_MZB/TaXon_tables/Sicily_eDNA_MZB_taxon_table_renamed_cons_derep_no_match_excluded_blanks_excluded_species.xlsx"
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna("nan")
    ## get a unique list of taxa
    taxa = set([taxon for taxon in TaXon_table_df["Species"].values.tolist() if taxon != "nan"])
    ## create a dataframe to store the results
    ## select only countries for the selected continents
    selected_countries_list = [country for country, values in country_codes_dict.items() if values[1] in continents_to_check]
    occurrence_df = pd.DataFrame(selected_countries_list, columns=["Country"])

    ## calculate runtime
    n_countries = len(selected_countries_list)
    n_species = len(taxa)
    t_single_request = 0.15
    t_total = round(t_single_request * n_species * n_countries / 60, 1)
    ## ask to continue the script
    answer = sg.PopupOKCancel("This will take roughly " + str(t_total) + " minutes. Continue?", title="Runtime")

    if answer == 'OK':

        ## create a subfolder for better sorting and overview
        dirName = Path(str(path_to_outdirs) + "/" + "Occurrence_analysis" + "/" + TaXon_table_xlsx.stem + "/")
        if not os.path.exists(dirName):
            os.mkdir(dirName)

        ############################################################################
        ## create the progress bar window
        layout = [[sg.Text('Progress bar')],
                  [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
                  [sg.Cancel()]]
        window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
        progress_bar = window_progress_bar['progressbar']
        progress_update = 0
        progress_increase = 1000 / len(taxa) + 1
        ############################################################################

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += 0
        progress_bar.UpdateBar(progress_update)
        ############################################################################

        ## request gbif for the occurrence data
        for taxon_name in taxa:
            occurrence_list= []
            for country, values in country_codes_dict.items():
                country_code = values[0]
                continent = values[1]
                ## only check selected continents to reduce runtime
                if continent in continents_to_check:
                    ## create an html session
                    with requests_html.HTMLSession() as session:
                        ## generate html request name
                        request_name = '%20'.join(taxon_name.split(' '))
                        ## request that name
                        r = session.get("https://api.gbif.org/v1/occurrence/search?scientificName=" + request_name + "&country=" + country_code)
                        ## parse json
                        res = json.loads(r.text)
                        ## get number of occurrences
                        occurrence_list.append(res["count"])
            ## store the results in the dataframe
            occurrence_df[taxon_name] = occurrence_list

            ############################################################################
            event, values = window_progress_bar.read(timeout=10)
            if event == 'Cancel'  or event is None:
                window_progress_bar.Close()
                raise RuntimeError
            # update bar with loop value +1 so that bar eventually reaches the maximum
            progress_update += progress_increase
            progress_bar.UpdateBar(progress_update)
            ############################################################################

        window_progress_bar.Close()

        ## remove countries that have 0 hits
        occurrence_df_filtered_list = []
        for row in occurrence_df.values.tolist():
            occurrences = set(row[1:])
            if occurrences != {0}:
                occurrence_df_filtered_list.append(row)

        ## create a dataframe with relative values
        occurrence_df_filtered_relative = pd.DataFrame(occurrence_df_filtered_list)
        occurrence_df_filtered_relative.columns = occurrence_df.columns.tolist()

        ## create a dataframe with absolute values
        occurrence_df_filtered_absolute = pd.DataFrame(occurrence_df_filtered_list)
        occurrence_df_filtered_absolute.columns = occurrence_df.columns.tolist()

        ## convert dataframe to relative occurrence abundance
        for taxon in taxa:
            df = occurrence_df_filtered_relative[["Country", taxon]]
            df_2 = df[taxon]/df[taxon].sum()
            df = df.assign(perc=df_2.values * 100)
            df = df.drop([taxon], axis=1)
            df = df.rename(columns={"perc": taxon})
            occurrence_df_filtered_relative[taxon] = df[taxon]

        fig = go.Figure()
        for row in occurrence_df_filtered_relative.values.tolist():
            occurrences = row[1:]
            country = row[0]
            fig.add_trace(go.Bar(x=list(taxa), y=list(occurrences), text=country, name=country, textposition='auto'))
        fig.update_layout(barmode='stack', width=int(width), height=int(height), template=template)
        fig.update_yaxes(title="GBIF occurrence references (%)")

        n_occurrences = []
        for taxon in taxa:
            n_occurrences.append(len([value for value in occurrence_df_filtered_relative[taxon].values.tolist() if value != 0]))

        fig.add_trace(go.Scatter(x=list(taxa), y=[105]*len(taxa), text=n_occurrences, name="countries", mode="text"))

        ## define output files
        output_pdf = Path(str(dirName) + "/" + '_'.join(continents_to_check) + ".pdf")
        output_html = Path(str(dirName) + "/" + '_'.join(continents_to_check) + ".html")
        output_xlsx = Path(str(dirName) + "/" + '_'.join(continents_to_check) + ".xlsx")

        ## write to different sheets, one for absolute data, one for relative
        with pd.ExcelWriter(output_xlsx) as writer:
            occurrence_df_filtered_relative.to_excel(writer, sheet_name='relative', index=False)
            occurrence_df_filtered_absolute.to_excel(writer, sheet_name='absolute', index=False)

        ## write figures
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## print closing text
        closing_text = "GBIF occurrence plots and tables are found under: " + str(path_to_outdirs) + "/Occurrence_analysis/"
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write log file
        from taxontabletools.create_log import ttt_log
        ttt_log("occurrence analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, "", path_to_outdirs)

def gbif_occurrence_plot(width, height, template, theme, path_to_outdirs):

    import requests_html, json, glob
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path
    import plotly.graph_objects as go
    import os, webbrowser

    # slice function for lists to split up lists
    def slices(list, slice):
        for i in range(0, len(list), slice):
            yield list[i : i + slice]

    ## collect all occurrence tables from the directory
    input_files = sorted(glob.glob(str(path_to_outdirs) + "/Occurrence_analysis/*/*.xlsx"))
    input_files_short = []
    for file in input_files:
        if "\\" in file:
            input_files_short.append(str(Path(file).parent).split("\\")[-1] + "/" + Path(file).name)
        else:
            input_files_short.append(str(Path(file).parent).split("/")[-1] + "/" + Path(file).name)

    ##################################################
    # start a second window to ask for the read tables to process

    win2_active = True
    input_files_list = list(slices([sg.Radio(name, "occurrence_tables", default=True) for name in sorted(input_files_short)], 2))
    layout2 = [[sg.Text("Occurrence plots", size=(20,1))],
    [sg.Frame(layout = input_files_list, title = 'Check occurrence table to re-plot')],
    [sg.Button('Plot')],
    [sg.Button('Back')]]

    win2 = sg.Window('Occurrence plots', layout2, keep_on_top=True)
    while True:
        event2, values2 = win2.Read()

        if event2 is None or event2 == 'Back':
            win2.Close()
            win2_active = False
            break

        if event2 == 'Plot':

            if values2 == {}:
                sg.Popup("Please first download occurrence data!")

            else:
                if len(values2) == 1:
                    file = Path(str(path_to_outdirs) + "/Occurrence_analysis/" + input_files_short[0])
                else:
                    file = Path(str(path_to_outdirs) + "/Occurrence_analysis/" + [input_files_short[value] for key, value in values2.items() if key == True][0])
                occurrence_df_filtered = pd.read_excel(file, sheet_name="relative").fillna(0)
                taxa = occurrence_df_filtered.columns.tolist()[1:]

                fig = go.Figure()
                for row in occurrence_df_filtered.values.tolist():
                    occurrences = row[1:]
                    country = row[0]
                    fig.add_trace(go.Bar(x=list(taxa), y=list(occurrences), text=country, name=country, textposition='auto'))
                fig.update_layout(barmode='stack', width=int(width), height=int(height), template=template)
                fig.update_yaxes(title="GBIF occurrence references (%)")

                n_occurrences = []
                for taxon in taxa:
                    n_occurrences.append(len([value for value in occurrence_df_filtered[taxon].values.tolist() if value != 0]))

                fig.add_trace(go.Scatter(x=taxa, y=[105]*len(taxa), text=n_occurrences, name="countries", mode="text"))

                ## define output files
                output_pdf = Path(str(file).replace(".xlsx", ".pdf"))
                output_html = Path(str(file).replace(".xlsx", ".html"))
                fig.write_image(str(output_pdf))
                fig.write_html(str(output_html))

                ## ask to show file
                answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
                if answer == "Yes":
                    webbrowser.open('file://' + str(output_html))

                ## print closing text
                closing_text = "GBIF occurrence plots and tables are found under: " + str(path_to_outdirs) + "/Occurrence_analysis/"
                sg.Popup(closing_text, title="Finished", keep_on_top=True)

                win2.Close()
                win2_active = False
                break



def create_taxon_table_per_sample(TaXon_table_xlsx, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path

    TaXon_table_file =  Path(TaXon_table_xlsx)

    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_file)
    TaXon_datasheet = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)
    TaXon_table = TaXon_datasheet.values.tolist()
    samples_to_process = TaXon_datasheet.columns[10:]
    first_ten_columns_header = TaXon_datasheet.columns[:10].values.tolist()
    first_ten_columns = TaXon_datasheet.iloc[:,[0,1,2,3,4,5,6,7,8,9]].values.tolist()
    OTU_list = TaXon_datasheet['IDs'].values.tolist()

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(samples_to_process) + 1
    ############################################################################

    for sample in samples_to_process:

        Output_name = Path(sample + ".xlsx")
        Output_file = path_to_outdirs / "TaXon_tables_per_sample" / Output_name

        read_numbers = TaXon_datasheet[sample].values.tolist()
        sample_rows_list = []

        for i, read_number in enumerate(read_numbers):
            if read_number > 0:
                sample_rows_list.append(first_ten_columns[i] + [read_number])

        headers_df = pd.DataFrame([first_ten_columns_header + [sample]])
        sample_df = pd.DataFrame(sample_rows_list)
        sample_df = headers_df.append(sample_df)

        sample_df.to_excel(Output_file, engine='xlsxwriter', sheet_name='TaXon table', index=False, header=False)

        ############################################################################
        event, values = window_progress_bar.read(timeout=10)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    closing_text = "\n" + "Taxon tables are found in: " + str(path_to_outdirs) + "/TaXon_tables_per_sample/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    placeholder = TaXon_table_file.name + " (multiple files)"
    ttt_log("taXon table per sample", "analysis", TaXon_table_file.name, placeholder, "nan", path_to_outdirs)



# krona chart
def create_krona_chart(TaXon_table_xlsx, path_to_outdirs):

    import subprocess, os, webbrowser
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path

    try:
        subprocess.call(["ktImportText"], stdout=open(os.devnull, 'wb'))
    except:
        sg.PopupError("Krona tools must be manually installed first!" + "\n" * 2 + "Note: Krona tools is currently not supported on Windows!" + "\n", title="Error")
        raise RuntimeError("Krona tools needs to be installed")

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    TaXon_table_df = TaXon_table_df.replace(np.nan, '__', regex=True)

    # check for presence absence data
    # otherwise abort and print error message
    pa_test = set([val for sublist in TaXon_table_df[TaXon_table_samples].values.tolist() for val in sublist])
    if pa_test == {1,0}:
        pa_data = True
    else:
        pa_data = False

    row1 = ["sample-ID", "", "", "", "", "", ""]
    row2 = ["count", "phylum", "class", "order", "family", "genus", "species"]
    krona_taxonomy_list = []
    krona_taxonomy_list.append(row1)
    krona_taxonomy_list.append(row2)

    for OTU in TaXon_table_df.values.tolist():
        taxonomy = OTU[1:7]
        reads = sum(OTU[10:])
        if pa_data == True:
            krona_taxonomy_list.append([1] + taxonomy)
        else:
            krona_taxonomy_list.append([reads] + taxonomy)

    krona_taxonomy_df = pd.DataFrame(krona_taxonomy_list)

    krona_chart_directory = Path(str(path_to_outdirs) + "/" + "Krona_charts" + "/" + TaXon_table_xlsx.stem)
    krona_table_tsv = Path(str(krona_chart_directory) + "_krona_table.tsv")
    krona_chart_html = Path(str(krona_chart_directory) + "_krona_table.html")

    # write krona table to tsv
    krona_taxonomy_df.to_csv(krona_table_tsv, sep="\t", header=False, index=False)

    os.system("ktImportText " + str(krona_table_tsv) + " -o " + str(krona_chart_html))

    # finish script
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(krona_chart_html))

    closing_text = "Krona chart is found under:\n" + '/'.join(str(krona_chart_html).split("/")[-4:])
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("krona chart", "analysis", TaXon_table_xlsx.name, krona_chart_html.name, "nan", path_to_outdirs)
